function ImgLoader() {
	"use strict";
	
	this.onprogress = function (loaded, total) {};
	this.oncompletion = function (media) {};
	
	this.loaded = this.total = 0;
	
	this.load = function (media) {
		var loader = this;
		for (var i in media) {
			if (typeof media[i] != "string") continue;
			this.total++;
			var fn = media[i];
			media[i] = new Image();
			media[i].src = fn;
			media[i].onload = function () {
				loader.loaded++;
				loader.onprogress(loader.loaded, loader.total);
				if (loader.loaded == loader.total)
					loader.oncompletion(media);
			}
		}
	}
}

function Game(canvas) {
	"use strict";
	
	this.canvas = canvas;
	this.context = canvas.getContext("2d");
	
	this.media = {};
	this.loaded = false;
	
	this.compass = ["n", "ne", "e", "se", "s", "sw", "w", "nw"];
	this.stepSize = {
		"n": {x: 0, y: 1},
		"ne": {x: 1, y: 1},
		"e": {x: 1, y: 0},
		"se": {x: 1, y: -1},
		"s": {x: 0, y: -1},
		"sw": {x: -1, y: -1},
		"w": {x: -1, y: 0},
		"nw": {x: -1, y: 1},
	};
	
	this.map_data = {
		"0,0": {sprite: "home", x: 0, y: 0},
		"-2,-2": {sprite: "column", x: -2, y: -2},
		"2,2": {sprite: "column", x: 2, y: 2},
	};

	this.camera = new EightWay.Camera(0, -3, -6, 500);
	this.scene = new EightWay.Scene();
	this.scene.addSprite("cursor", 0, 0, -1);
	this.scene.viewDist = 3;

	this.cursor = this.scene.sprites[0];
	this.facing = 0;
	this.pedometer = 0;
	this.zoomDist = this.camera.y;
	this.freeWalk = true;
	this.editing = true;

	this.loadMedia = function () {
		var loader = new ImgLoader();
		loader.onprogress = (function (loaded, total) {
			this.render_loading(loaded, total);
		}).bind(this);
		loader.oncompletion = (function (media) {
			this.loaded = true;
			this.media = media;
			this.render();
		}).bind(this);
		loader.load({
			bg1: "art/bg1n.png",
			bg2: "art/bg2n.png",
			column: "art/column.png",
			column2: "art/column2.png",
			home: "art/teleporter-on.png",
		});
	};
	
	this.render_loading = function (loaded, total) {
		var ctx = this.context;
		ctx.save();

		ctx.scale(this.canvas.width / 800, this.canvas.height / 500);

		ctx.fillStyle = "rgb(20, 12, 28)";
		ctx.fillRect(0, 0, 800, 500);

		ctx.font = "75px monospace";
		ctx.fillStyle = "#d04648";
		ctx.textAlign = "center";
		ctx.textBaseline = "middle";
		ctx.fillText("Loading " + loaded + " of " + total, 400, 250);
		
		ctx.restore();
	};

	this.resize = function () {
		if (this.loaded)
			this.render();
	};
	
	this.render = function () {
		var ctx = this.context;
		
		ctx.save();
		ctx.scale(this.canvas.width / 800, this.canvas.height / 500);

		if (this.editing) {
			ctx.fillStyle = "rgb(20, 12, 28)";
			ctx.fillRect(0, 0, 800, 500);
		} else if (this.facing % 2 == 0) {
			ctx.drawImage(this.media["bg1"], 0, 0, 800, 500);
		} else {
			ctx.drawImage(this.media["bg2"], 0, 0, 800, 500);
		}
		
		ctx.save();
		if (this.editing)
			ctx.translate(400, 0);
		else
			ctx.translate(400, 250);

		this.scene.sprites.length = 1;
		for (var i in this.map_data) {
			var s = this.map_data[i];
			this.scene.addSprite(s.sprite, s.x, s.y);
		}
		if (this.editing)
			this.addGrid();
		this.scene.render(this.camera, this.compass[this.facing]);
		for (var i = 0; i < this.scene.visible.length; i++)
			this.renderSprite(this.scene.visible[i]);
		ctx.restore();

		ctx.fillStyle = "rgb(109, 170, 44)";
		ctx.textBaseline = "bottom";

		ctx.textAlign = "left";
		ctx.font = "bold 50px monospace";
		var text = this.compass[this.facing].toUpperCase();
		ctx.fillText(text, 0, 50);
		ctx.font = "bold 25px monospace";
		text = this.scene.center.x + "/" + this.scene.center.y;
		ctx.fillText(text, 0, 75);
		
		ctx.textAlign = "right";
		ctx.font = "bold 50px monospace";
		if (this.editing) {
			ctx.fillText("EDIT", 800, 50);
		} else {
			if (this.freeWalk)
				ctx.fillText("WALK FREE", 800, 50);
			else
				ctx.fillText("WALK", 800, 50);
			ctx.font = "bold 25px monospace";
			text = "steps: " + this.pedometer;
			ctx.fillText(text, 800, 75);
		}
		
		ctx.restore();
	};

	this.renderCursor = function (x, y, scale) {
		var ctx = this.context;
		ctx.strokeStyle = "rgb(109, 170, 44)";
		ctx.lineWidth = scale / 16;
		ctx.beginPath();
		ctx.moveTo(x, y);
		ctx.lineTo(x - scale / 4, y - scale / 2);
		ctx.lineTo(x + scale / 4, y - scale / 2);
		ctx.closePath();
		ctx.stroke();
	};
	
	this.renderSprite = function (sprite) {
		if (sprite.id === "cursor") {
			this.renderCursor(sprite.x, sprite.y, sprite.scale);
		} else if (sprite.id === "grid") {
			this.context.fillStyle = "rgb(109, 170, 44)";
			this.context.fillRect(
				sprite.x - sprite.scale / 32,
				sprite.y - sprite.scale / 32,
				sprite.scale / 16, sprite.scale / 32);
		} else {
			this.context.drawImage(
				this.media[sprite.id],
				sprite.x - sprite.scale * 0.5,
				sprite.y - sprite.scale + sprite.scale / 10,
				sprite.scale, sprite.scale);
		}
	};
	
	this.addGrid = function () {
		var c = this.scene.center;
		var d = this.scene.viewDist;
		for (var y = c.y - d; y <= c.y + d; y++)
			for (var x = c.x - d; x <= c.x + d; x++)
				if (!((x + "," + y) in this.map_data))
					this.scene.addSprite("grid", x, y);
	};
	
	this.canEnter = function (x, y) {
		if (this.editing || this.freeWalk) {
			return true;
		} else {
			var key = x + "," + y;
			if (!(key in this.map_data))
				return true;
			else if (this.map_data[key].sprite === "home")
				return true;
			else
				return false;
		}
	};
	
	this.handleCommand = function (command) {
		if (command === "left") {
			this.facing--;
			if (this.facing < 0)
				this.facing = 7;
		} else if (command === "right") {
			this.facing = (this.facing + 1) % 8;
		} else if (command === "up") {
			var dir = this.compass[this.facing];
			var delta = this.stepSize[dir];
			var new_x = this.scene.center.x + delta.x;
			var new_y = this.scene.center.y + delta.y;
			if (this.canEnter(new_x, new_y)) {
				this.scene.center.x += delta.x;
				this.scene.center.y += delta.y;
				if (this.editing) {
					this.cursor.x = this.scene.center.x;
					this.cursor.y = this.scene.center.y;
				} else {
					this.pedometer++;
				}
			}
		} else if (command === "down") {
			var dir = this.compass[this.facing];
			var delta = this.stepSize[dir];
			var new_x = this.scene.center.x - delta.x;
			var new_y = this.scene.center.y - delta.y;
			if (this.canEnter(new_x, new_y)) {
				this.scene.center.x -= delta.x;
				this.scene.center.y -= delta.y;
				if (this.editing) {
					this.cursor.x = this.scene.center.x;
					this.cursor.y = this.scene.center.y;
				} else {
					this.pedometer++;
				}
			}
		} else if (command === "menu") {
			this.editing = !this.editing;
			if (this.editing) {
				this.scene.center.x = this.cursor.x;
				this.scene.center.y = this.cursor.y;
				this.camera.y = this.zoomDist;
				this.camera.z = this.camera.y * 2;
				this.scene.viewDist = -this.camera.y;
				if (this.scene.spacing < 1)
					this.scene.viewDist *= 2;
			} else {
				this.camera.y = -0.5;
				this.camera.z = -0.5;
				this.scene.viewDist = 15;
				this.pedometer = 0;
			}
		} else if (command === "zoomIn") {
			if (!this.editing) {
				if (this.scene.spacing.u === 1)
					this.scene.spacing.u = 0.6;
				else
					this.scene.spacing.u = 1;
			} else if (this.camera.y < -2) {
				this.camera.y++;
				this.camera.z = this.camera.y * 2;
				this.scene.viewDist = -this.camera.y;
				if (this.scene.spacing < 1)
					this.scene.viewDist *= 2;
				this.zoomDist = this.camera.y;
			}
		} else if (command === "zoomOut") {
			if (!this.editing) {
				if (this.scene.spacing.u === 1)
					this.scene.spacing.u = 0.6;
				else
					this.scene.spacing.u = 1;
			} else {
				this.camera.y--;
				this.camera.z = this.camera.y * 2;
				this.scene.viewDist = -this.camera.y;
				if (this.scene.spacing < 1)
					this.scene.viewDist *= 2;
				this.zoomDist = this.camera.y;
			}
		} else if (command === "action") {
			if (!this.editing) {
				this.freeWalk = !this.freeWalk;
			} else {
				var c = this.cursor;
				var key = c.x + "," + c.y;
				if (c.x === 0 && c.y === 0) {
					// Do nothing.
				} else if (key in this.map_data) {
					delete this.map_data[key];
				} else if ((c.x + c.y) % 2 === 0) {
					this.map_data[key] = {
						sprite: "column",
						x: c.x,
						y: c.y
					};
				} else {
					this.map_data[key] = {
						sprite: "column2",
						x: c.x,
						y: c.y
					};
				}
			}
		};
		this.render();
	};
}
